from collections import defaultdict
from tqdm import tqdm
import evaluate
from datasets import load_from_disk


def ner_seq_from(chars, ners):
    pred = ["O" for _ in range(len(chars))]

    for ner_index in ners:
        pred[ner_index] = "Sep"

    return pred


def try_fix_last_index(text, indices):
    last_index = indices[-1]

    if last_index >= len(text):
        indices[-1] = last_index - 1


def model_chonky(model_id, **kwargs):
    from src.chonky import ParagraphSplitter

    splitter = ParagraphSplitter(
        model_id=model_id,
        device="cuda",
        **kwargs,
    )

    def predict(x):
        last_char_idx = 0
        pred_char_indices = []

        for chunk in splitter(x):
            last_char_idx += len(chunk)

            pred_char_indices.append(last_char_idx)

        try_fix_last_index(x, pred_char_indices)

        return ner_seq_from(x, pred_char_indices)

    return predict


def model_chonkie_semantic(embedding_model):
    from chonkie import SemanticChunker

    chunker = SemanticChunker(embedding_model=embedding_model)

    def predict(x):
        chunks = chunker(x)

        pred_char_indices = [chunk.end_index for chunk in chunks]
        try_fix_last_index(x, pred_char_indices)

        return ner_seq_from(x, pred_char_indices)

    return predict


def model_chonkie_recursive():
    from chonkie import RecursiveChunker

    chunker = RecursiveChunker()

    def predict(x):
        chunks = chunker(x)

        pred_char_indices = [chunk.end_index for chunk in chunks]
        try_fix_last_index(x, pred_char_indices)

        return ner_seq_from(x, pred_char_indices)

    return predict


def model_sat(model_id, do_paragraph_segmentation):
    from wtpsplit import SaT

    model = SaT(model_id)
    model.to("cuda")

    def predict(x):
        last_char_idx = 0
        pred_char_indices = []

        for sents in model.split(
            x, do_paragraph_segmentation=do_paragraph_segmentation
        ):
            if not do_paragraph_segmentation:
                sents = [sents]

            last_char_idx += sum(map(len, sents))

            pred_char_indices.append(last_char_idx - 1)

        return ner_seq_from(x, pred_char_indices)

    return predict


def model_llama_index_semantic_splitter(embedding_model):
    from llama_index.core.schema import Document
    from llama_index.core.node_parser import SemanticSplitterNodeParser
    from llama_index.embeddings.huggingface import HuggingFaceEmbedding

    splitter = SemanticSplitterNodeParser(
        embed_model=HuggingFaceEmbedding(model_name=embedding_model)
    )

    def predict(x):
        doc = Document(text=x)
        nodes = splitter.get_nodes_from_documents([doc])
        pred_char_indices = [node.end_char_idx - 1 for node in nodes]
        try_fix_last_index(x, pred_char_indices)

        return ner_seq_from(x, pred_char_indices)

    return predict


def model_langchain_semantic_chunker(embedding_model):
    from langchain_experimental.text_splitter import SemanticChunker
    from langchain_huggingface import HuggingFaceEmbeddings

    hf_embeddings = HuggingFaceEmbeddings(model_name=embedding_model)
    splitter = SemanticChunker(hf_embeddings)

    def predict(x):
        docs = splitter.create_documents([x])

        last_char_idx = 0
        pred_char_indices = []

        for doc in docs:
            last_char_idx += len(doc.page_content)

            pred_char_indices.append(last_char_idx + 1)

        try_fix_last_index(x, pred_char_indices)

        return ner_seq_from(x, pred_char_indices)

    return predict


def model_langchain_recursive_chunker():
    from langchain_text_splitters import RecursiveCharacterTextSplitter

    splitter = RecursiveCharacterTextSplitter(chunk_overlap=0)

    def predict(x):
        docs = splitter.create_documents([x])

        last_char_idx = 0
        pred_char_indices = []

        for doc in docs:
            last_char_idx += len(doc.page_content)

            pred_char_indices.append(last_char_idx + 1)

        try_fix_last_index(x, pred_char_indices)

        return ner_seq_from(x, pred_char_indices)

    return predict


def make_gt(tokens, ner_tags):
    char_indices = []
    char_it = 0
    for token, tag in zip(tokens, ner_tags):
        char_it += len(token) + 1

        if tag:
            char_indices.append(char_it - 1)

    char_indices[-1] -= 1

    return ner_seq_from(" ".join(tokens), char_indices)


def eval_loop(eval_dataset, models):
    all_outputs = defaultdict(list)
    gts = []

    assert len(eval_dataset["tokens"]) == len(eval_dataset["ner_tags"])

    tokens = eval_dataset["tokens"]
    ner_tags = eval_dataset["ner_tags"]

    outputs_to_assert = []
    for model_name, model in tqdm(models):
        text = " ".join(tokens)

        output = model(text)
        all_outputs[model_name].append(output)
        outputs_to_assert.append(output)

    gt = make_gt(tokens, ner_tags)
    gts.append(gt)

    assert len(set(list(map(len, outputs_to_assert)) + [len(gt)])) == 1

    return all_outputs, gts


def pretty_print_metrics(all_metrics):
    from operator import itemgetter
    from tabulate import tabulate

    headers = ["Model"]
    rows = []

    is_header_set = False
    by_model = itemgetter(0)
    all_metrics = sorted(all_metrics.items(), key=by_model)
    for model_name, for_model in all_metrics:
        by_dataset = itemgetter(0)
        for_model = sorted(for_model, key=by_dataset)

        row = []
        for dataset_name, metrics in for_model:
            if not is_header_set:
                headers.append(dataset_name)

            metric_value = round(metrics["overall_f1"], 2)
            row.append(metric_value)

        row = [model_name] + row
        rows.append(row)
        is_header_set = True

    print(tabulate(rows, headers=headers, tablefmt="github"))


def main():
    dataset_names = [
        "bookcorpus",
        "en_judgements",
        "paul_graham",
        "20_newsgroups",
    ]

    models = [
        (
            "SaT(sat-12l-sm, do_ps=True)",
            model_sat("sat-12l-sm", do_paragraph_segmentation=True),
        ),
        (
            "SaT(sat-12l-sm, do_ps=False)",
            model_sat("sat-12l-sm", do_paragraph_segmentation=False),
        ),
        ("SaT sat-3l do_ps=True", model_sat("sat-3l", do_paragraph_segmentation=True)),
        (
            "SaT(sat-3l, do_ps=False)",
            model_sat("sat-3l", do_paragraph_segmentation=False),
        ),
        (
            "chonkY_distilbert",
            model_chonky(model_id="mirth/chonky_distilbert_uncased_1"),
        ),
        (
            "chonkY_modernbert",
            model_chonky(
                model_id="mirth/chonky_modernbert_base_1",
                _attn_implementation="sdpa",
                reference_compile=False,
            ),
        ),
        (
            "chonkIE SemanticChunker(potion-base-8M)",
            model_chonkie_semantic(embedding_model="minishlab/potion-base-8M"),
        ),
        (
            "chonkIE SemanticChunker(bge-small-en-v1.5)",
            model_chonkie_semantic(embedding_model="BAAI/bge-small-en-v1.5"),
        ),
        ("chonkIE RecursiveChunker", model_chonkie_recursive()),
        (
            "llamaindex SemanticSplitter(bge-small-en-v1.5)",
            model_llama_index_semantic_splitter(
                embedding_model="BAAI/bge-small-en-v1.5"
            ),
        ),
        (
            "langchain SemanticChunker(all-mpnet-base-v2)",
            model_langchain_semantic_chunker(
                embedding_model="sentence-transformers/all-mpnet-base-v2"
            ),
        ),
        (
            "langchain SemanticChunker(bge-small-en-v1.5)",
            model_langchain_semantic_chunker(embedding_model="BAAI/bge-small-en-v1.5"),
        ),
        (
            "langchain SemanticChunker(potion-base-8M)",
            model_langchain_semantic_chunker(
                embedding_model="minishlab/potion-base-8M"
            ),
        ),
        ("langchain RecursiveChar", model_langchain_recursive_chunker()),
    ]

    all_metrics = defaultdict(list)
    for dataset_name in dataset_names:
        eval_dataset = load_from_disk(f"data/{dataset_name}")
        eval_dataset = eval_dataset[:1000000]

        results, gts = eval_loop(eval_dataset, models)

        seqeval = evaluate.load("seqeval")

        for model_name, preds in sorted(results.items(), key=lambda pair: pair[0]):
            metrics = seqeval.compute(predictions=preds, references=gts)
            all_metrics[model_name].append((dataset_name, metrics))

    pretty_print_metrics(all_metrics)


if __name__ == "__main__":
    main()



from operator import itemgetter
from collections import defaultdict
import numpy as np
import pandas as pd
from datasets import Dataset


def dataset_refined_bookcorpus():
    from datasets import load_from_disk

    ds = load_from_disk("data/refined-bookcorpus-dataset_hf_test")

    all_paragraphs = []
    for tokens, ner_tags in zip(ds["tokens"], ds["ner_tags"]):
        prev_tag_index = 0

        for tag_index in np.nonzero(ner_tags)[0]:
            tag_index += 1
            para = " ".join(tokens[prev_tag_index:tag_index])
            all_paragraphs.append(para)

            prev_tag_index = tag_index

    return all_paragraphs


def dataset_MultiLegalSBD_en_judgements():
    filenames = [
        "data/MultiLegalSBD/data/en_judgements_0.jsonl",
        "data/MultiLegalSBD/data/en_judgements_1.jsonl",
        "data/MultiLegalSBD/data/en_judgements_2.jsonl",
        "data/MultiLegalSBD/data/en_judgements_3.jsonl",
    ]

    dfs = []
    for filename in filenames:
        dfs.append(pd.read_json(filename, lines=True))

    df = pd.concat(dfs)

    all_paragraphs = []

    for text in df.text.values:
        paragraphs = list(filter(len, text.split("\r\n")))

        all_paragraphs.extend(paragraphs)

    return all_paragraphs


def hf_dataset_with_paragraphs(dataset_id, split):
    from datasets import load_dataset

    ds = load_dataset(dataset_id)

    all_paragraphs = []

    for text in ds[split]["text"]:
        paragraphs = list(filter(len, text.split("\n")))

        all_paragraphs.extend(paragraphs)

    return all_paragraphs


def to_dict(paragraphs):
    tokens = []
    ner_tags = []

    for para in paragraphs:
        for token in para.split(" "):
            tokens.append(token)
            ner_tags.append(0)

        ner_tags[-1] = 1

    assert len(tokens) == len(ner_tags)

    return {"tokens": tokens, "ner_tags": ner_tags}


def main():
    raw_datasets = [
        ("bookcorpus", dataset_refined_bookcorpus()),
        ("en_judgements", dataset_MultiLegalSBD_en_judgements()),
        (
            "paul_graham",
            hf_dataset_with_paragraphs(
                dataset_id="sgoel9/paul_graham_essays", split="train"
            ),
        ),
        (
            "20_newsgroups",
            hf_dataset_with_paragraphs(dataset_id="SetFit/20_newsgroups", split="test"),
        ),
    ]

    for dataset_name, paragraphs in raw_datasets:
        dataset = Dataset.from_dict(to_dict(paragraphs))
        dataset.save_to_disk(f"data/{dataset_name}")


if __name__ == "__main__":
    main()



import numpy as np
import torch
import evaluate
from datasets import load_from_disk
from transformers import (
    AutoTokenizer,
    AutoModelForTokenClassification,
    TrainingArguments,
    Trainer,
    DataCollatorForTokenClassification,
)

torch.set_float32_matmul_precision("high")


def distilbert(model_name):
    id2label = {
        0: "O",
        1: "separator",
    }
    label2id = {
        "O": 0,
        "separator": 1,
    }

    model = AutoModelForTokenClassification.from_pretrained(
        model_name, num_labels=2, id2label=id2label, label2id=label2id
    )

    tokenizer = AutoTokenizer.from_pretrained(model_name)

    return model, tokenizer


def modernbert(model_id):
    id2label = {
        0: "O",
        1: "separator",
    }
    label2id = {
        "O": 0,
        "separator": 1,
    }

    model = AutoModelForTokenClassification.from_pretrained(
        model_id,
        num_labels=2,
        id2label=id2label,
        label2id=label2id,
        _attn_implementation="flash_attention_2",
    )

    tokenizer = AutoTokenizer.from_pretrained(model_id)

    return model, tokenizer


def main(dataset_id, model_id, output_dir, batch_size, max_seq_len=None):
    model, tokenizer = modernbert(model_id=model_id)
    # model, tokenizer = distilbert(model_name=model_id)

    if max_seq_len is None:
        max_seq_len = tokenizer.model_max_length

    dataset = load_from_disk(dataset_id)
    print(dataset)
    dataset_val = dataset["test"]
    dataset_train = dataset["train"]
    label_list = ["O", "separator"]
    seqeval = evaluate.load("seqeval")

    def tokenize_and_align_labels(examples):
        tokenized_inputs = tokenizer(
            examples["tokens"],
            truncation=True,
            is_split_into_words=True,
            max_length=max_seq_len,
        )

        labels = []
        for i, label in enumerate(examples["ner_tags"]):
            word_ids = tokenized_inputs.word_ids(
                batch_index=i
            )  # Map tokens to their respective word.
            previous_word_idx = None
            label_ids = []
            for word_idx in word_ids:  # Set the special tokens to -100.
                if word_idx is None:
                    label_ids.append(-100)
                elif (
                    word_idx != previous_word_idx
                ):  # Only label the first token of a given word.
                    label_ids.append(label[word_idx])
                else:
                    label_ids.append(-100)
                previous_word_idx = word_idx
            labels.append(label_ids)

        tokenized_inputs["labels"] = labels

        return tokenized_inputs

    def compute_metrics(p):
        predictions, labels = p
        predictions = np.argmax(predictions, axis=2)

        true_predictions = [
            [label_list[p] for (p, l) in zip(prediction, label) if l != -100]
            for prediction, label in zip(predictions, labels)
        ]
        true_labels = [
            [label_list[l] for (p, l) in zip(prediction, label) if l != -100]
            for prediction, label in zip(predictions, labels)
        ]

        results = seqeval.compute(predictions=true_predictions, references=true_labels)
        return {
            "precision": results["overall_precision"],
            "recall": results["overall_recall"],
            "f1": results["overall_f1"],
            "accuracy": results["overall_accuracy"],
        }

    tokenized_dataset_train = dataset_train.map(tokenize_and_align_labels, batched=True)
    tokenized_dataset_val = dataset_val.map(tokenize_and_align_labels, batched=True)

    data_collator = DataCollatorForTokenClassification(tokenizer=tokenizer)
    training_args = TrainingArguments(
        output_dir=output_dir,
        learning_rate=2e-5,
        per_device_train_batch_size=batch_size,
        per_device_eval_batch_size=batch_size,
        num_train_epochs=25,
        weight_decay=0.01,
        eval_strategy="epoch",
        save_strategy="epoch",
        load_best_model_at_end=True,
        push_to_hub=False,
        fp16=True,
    )

    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=tokenized_dataset_train,
        eval_dataset=tokenized_dataset_val,
        processing_class=tokenizer,
        data_collator=data_collator,
        compute_metrics=compute_metrics,
    )

    trainer.train()


if __name__ == "__main__":
    main(
        dataset_id="data/refined-bookcorpus-dataset_hf_ModernBERT-base1024",
        model_id="modernbert_large_minipile/checkpoint-54717",
        output_dir="modernbert_large_minipile_plus_bookcorpus",
        batch_size=64,
        max_seq_len=1024,
    )



from transformers import AutoTokenizer, AutoModelForTokenClassification, pipeline


def split_into_semantic_chunks(text, ners):
    begin_index = 0

    for ner in ners:
        chunk = text[begin_index : ner["end"]]
        yield chunk
        begin_index = ner["end"]

    yield text[begin_index:]


class ParagraphSplitter:
    def __init__(
        self, model_id="mirth/chonky_distilbert_uncased_1", device="cpu", **model_kwargs
    ):
        if model_id in (
            "mirth/chonky_modernbert_base_1",
            "mirth/chonky_modernbert_large_1",
        ):
            tokenizer_kwargs = {"model_max_length": 1024}
        else:
            tokenizer_kwargs = {}

        self.tokenizer = AutoTokenizer.from_pretrained(model_id, **tokenizer_kwargs)

        id2label = {
            0: "O",
            1: "separator",
        }
        label2id = {
            "O": 0,
            "separator": 1,
        }

        model = AutoModelForTokenClassification.from_pretrained(
            model_id,
            num_labels=2,
            id2label=id2label,
            label2id=label2id,
            **model_kwargs,
        )
        model.to(device)
        self.pipe = pipeline(
            "ner",
            model=model,
            tokenizer=self.tokenizer,
            device=device,
            aggregation_strategy="simple",
            stride=self.tokenizer.model_max_length // 2,
        )

    def __call__(self, text):
        output = self.pipe(text)

        yield from split_into_semantic_chunks(text, output)


if __name__ == "__main__":
    with open(
        "../../data/paul_graham_essay_no_new_line/paul_graham_essay_no_new_line.txt"
    ) as file:
        pg = file.read()

    c = ParagraphSplitter(device="cuda")
    for sem_chunk in c(pg):
        print(sem_chunk)
        print("--")
